@@ -36,8 +36,8 @@ Known limitations
   this is currently limited to cx_Oracle.
 """
 __author__ = "Andy Todd <andy47@halfcooked.com>"
-__date__ = (2009, 8, 13)
-__version__ = (0, 3, 0)
+__date__ = (2010, 1, 7)
+__version__ = (0, 3, 5)
 
 import schema
 import sys
@@ -48,7 +48,8 @@ TEXT_DATATYPES = ['CHAR', 'NCHAR', 'NVARCHAR2', 'VARCHAR2']
 NUMERIC_DATATYPES = ['FLOAT', 'NUMBER']
 DATE_DATATYPES = ['DATE', 'TIMESTAMP']
 DEFAULT_NUM_LENGTH = '38'
-
+SYSTEM_SCHEMAS = ['MDSYS', 'TSMSYS', 'OUTLN', 'CTXSYS', 'FLOWS_FILES', \
+                  'DBSNMP', 'XDB', 'SYS'] 
 
 class Schema(schema.Schema):
     """
@@ -72,6 +73,10 @@ class Schema(schema.Schema):
         """
         LOG.info('Getting details for Oracle schema %s from database' % self.name)
         schema = {}
+        # Need to know the Oracle schema that we are reading
+        stmt = "SELECT user FROM dual"
+        cursor.execute(stmt)
+        schema_name = cursor.fetchone()[0]
         # Tables
         stmt = """SELECT table_name 
                   FROM   user_tables 
@@ -79,21 +84,21 @@ class Schema(schema.Schema):
         cursor.execute(stmt)
         for table in cursor.fetchall():
             LOG.debug('Getting details for table %s' % table[0])
-            schema[table[0]] = Table(table[0], cursor)
+            schema[table[0]] = Table(table[0], cursor, schema_name)
         # Views
         stmt = """SELECT view_name
                   FROM   user_views"""
         cursor.execute(stmt)
         for view in cursor.fetchall():
             LOG.debug("Getting details for view %s" % view[0])
-            schema[view[0]] = View(view[0], cursor)
+            schema[view[0]] = View(view[0], cursor, schema_name)
         # Sequences
         stmt = """SELECT sequence_name
                   FROM   user_sequences"""
         cursor.execute(stmt)
         for sequence in cursor.fetchall():
             LOG.debug("Getting details for sequence %s" % sequence[0])
-            schema[sequence[0]] = Sequence(sequence[0], cursor)
+            schema[sequence[0]] = Sequence(sequence[0], cursor, schema_name)
         # Code objects (packages, procedures and functions)
         # Note that we exclude package bodies, even though they are separate
         # objects they are stored under the package header in gerald
@@ -107,9 +112,11 @@ class Schema(schema.Schema):
             object_name, object_type = code_object
             LOG.debug("Getting details for code object %s" % object_name)
             if object_type == 'PACKAGE':
-                schema[object_name] = Package(object_name, object_type, cursor)
+                schema[object_name] = Package(object_name, object_type, cursor,
+                        schema_name)
             else:
-                schema[object_name] = CodeObject(object_name, object_type, cursor)
+                schema[object_name] = CodeObject(object_name, object_type,
+                        cursor, schema_name)
         # Database links
         stmt = """SELECT db_link
                   FROM   user_db_links
@@ -125,6 +132,83 @@ class Schema(schema.Schema):
         return schema
 
 
+class User(schema.Schema):
+    """
+    A representation of an Oracle database user 
+    
+    This will contain all of the objects the supplied user can access.
+    """
+    def _get_schema(self, cursor):
+        """
+        Get definitions for all of the objects our user can see
+        
+        We query the data dictionary for (in order);
+          - Tables, Views, Sequences, Code objects, DB links
+        We should also get;
+          - Grants and synonyms
+
+        Note that object keys are 'schema.name' in this Class
+        Note also that we don't capture database links (yet)
+
+        @param cursor: The cursor to use to query the data dictionary
+        @type cursor: Database cursor object
+        @return: All of the objects in this schema
+        @rtype: Dictionary
+        """
+        LOG.info('Getting details for Oracle user %s from database' % self.name)
+        schema = {}
+        # Tables
+        stmt = """SELECT owner, table_name 
+                  FROM   all_tables 
+                  WHERE  table_name NOT LIKE 'DR%%'
+                  AND    owner NOT IN ('%s')
+               """ % ''','''.join(SYSTEM_SCHEMAS)
+        cursor.execute(stmt)
+        for table in cursor.fetchall():
+            owner, table_name = table
+            LOG.debug('Getting details for table %s' % table_name)
+            table_key = '%s.%s' % (owner, table_name)
+            schema[table_key] = Table(table_name, cursor, owner)
+        # Views
+        stmt = """SELECT view_name, owner
+                  FROM   all_views"""
+        cursor.execute(stmt)
+        for view in cursor.fetchall():
+            view_name, owner = view
+            LOG.debug("Getting details for view %s" % view_name)
+            view_key = '%s.%s' % (owner, view_name)
+            schema[view_key] = View(view_name, cursor, owner)
+        # Sequences
+        stmt = """SELECT sequence_name, sequence_owner
+                  FROM   all_sequences"""
+        cursor.execute(stmt)
+        for seq in cursor.fetchall():
+            sequence, owner = seq
+            LOG.debug("Getting details for sequence %s" % sequence)
+            sequence_key = '%s.%s' % (owner, sequence)
+            schema[sequence_key] = Sequence(sequence, cursor, owner)
+        # Code objects (packages, procedures and functions)
+        # Note that we exclude package bodies, even though they are separate
+        # objects they are stored under the package header in gerald
+        stmt = """SELECT object_name, object_type, owner
+                  FROM   all_objects
+                  WHERE  object_name IN (SELECT object_name FROM all_procedures)
+                  AND    object_type NOT IN ('PACKAGE BODY')
+               """
+        cursor.execute(stmt)
+        for code_object in cursor.fetchall():
+            object_name, object_type, owner = code_object
+            object_key = '%s.%s' % (owner, object_name)
+            LOG.debug("Getting details for code object %s" % object_key)
+            if object_type == 'PACKAGE':
+                schema[object_key] = Package(object_name, object_type, cursor, owner)
+            else:
+                schema[object_key] = CodeObject(object_name, object_type, cursor, owner)
+        # All done, return the fruit of our labours
+        LOG.info('Got details for user %s' % self.name)
+        return schema
+
+
 class OracleCalcPrecisionMixin(object):
     """Class to contain the calc_precision static method to be used as a 
     mixin by other classes in this module.
@@ -196,9 +280,10 @@ class Table(schema.Table, OracleCalcPrecisionMixin):
             uc_table_name = self.name.upper()
         # Table information
         stmt = """SELECT tablespace_name
-                  FROM   user_tables
-                  WHERE  table_name = :table_name""" 
-        cursor.execute(stmt, (uc_table_name, ))
+                  FROM   all_tables
+                  WHERE  table_name = :table_name
+                  AND    owner = :schema_name""" 
+        cursor.execute(stmt, (uc_table_name, self.schema))
         LOG.debug('Getting tablespace for table %s' % uc_table_name)
         result = cursor.fetchone()
         if result is None:
@@ -206,10 +291,11 @@ class Table(schema.Table, OracleCalcPrecisionMixin):
         self.tablespace_name = result[0]
         # Table Comment(s) - Note that Oracle only allows 1 per table
         stmt = """SELECT comments
-                  FROM   user_tab_comments
+                  FROM   all_tab_comments
                   WHERE  table_name = :table_name
+                  AND    owner = :schema_name
                   AND    comments IS NOT NULL"""
-        cursor.execute(stmt, (uc_table_name, ))
+        cursor.execute(stmt, (uc_table_name, self.schema))
         results = cursor.fetchall()
         if len(results) > 0: # Has the query returned at least 1 row?
             LOG.debug('Getting comments for table %s' % uc_table_name)
@@ -218,12 +304,14 @@ class Table(schema.Table, OracleCalcPrecisionMixin):
         stmt = """SELECT utc.column_id, utc.column_name, utc.data_type, 
                          utc.data_length, utc.data_precision, utc.data_scale,
                          utc.nullable, utc.data_default, ucc.comments
-                  FROM   user_tab_columns utc
-                         LEFT OUTER JOIN user_col_comments ucc ON 
+                  FROM   all_tab_columns utc
+                         LEFT OUTER JOIN all_col_comments ucc ON 
                            (utc.table_name = ucc.table_name AND
-                            utc.column_name = ucc.column_name)
-                  WHERE  utc.table_name = :table_name""" 
-        cursor.execute(stmt, (uc_table_name, ))
+                            utc.column_name = ucc.column_name AND
+                            utc.owner = ucc.owner)
+                  WHERE  utc.table_name = :table_name
+                  AND    utc.owner = :schema_name""" 
+        cursor.execute(stmt, (uc_table_name, self.schema))
         LOG.debug('Getting columns for table %s' % uc_table_name)
         for row in cursor.fetchall():
             # Column name is in the second element
@@ -255,54 +343,67 @@ class Table(schema.Table, OracleCalcPrecisionMixin):
                                   constraint_type,
                          r_constraint_name, search_condition,
                          decode(status, 'ENABLED', 'Y', 'DISABLED', 'N') status
-                  FROM   user_constraints
-                  WHERE  table_name = :table_name""" 
-        cursor.execute(stmt, (uc_table_name, ))
+                  FROM   all_constraints
+                  WHERE  table_name = :table_name
+                  AND    owner = :schema_name""" 
+        cursor.execute(stmt, (uc_table_name, self.schema))
         LOG.debug('Getting constraints for table %s' % uc_table_name)
         for key in cursor.fetchall():
             constraint = { 'name': key[0], 'type': key[1] }
+            LOG.debug('Getting details for %s which is a %s constraint' %
+                    (constraint['name'], constraint['type']))
             if key[4] == 'Y':
                 constraint['enabled'] = True
             else:
                 constraint['enabled'] = False
             if constraint['type'] in ('Foreign', 'Primary'):
                 stmt = """SELECT column_name
-                          FROM   user_cons_columns
+                          FROM   all_cons_columns
                           WHERE  table_name = :table_name
+                          AND    owner = :schema_name
                           AND    constraint_name = :constraint_name
                           ORDER BY position""" 
-                cursor.execute(stmt, (uc_table_name, constraint['name']))
+                cursor.execute(stmt, (uc_table_name, self.schema, constraint['name']))
                 columns = []
                 for column in cursor.fetchall():
                     columns.append(column[0])
                 constraint['columns'] = columns
                 if constraint['type'] == 'Foreign':
                     constraint['refpk'] = key[2]
+                    LOG.debug('Getting reference constraint %s' % constraint['refpk'])
                     stmt = """SELECT table_name
-                              FROM   user_constraints
+                              FROM   all_constraints
                               WHERE  constraint_name = :constraint_name""" 
+                    # Foreign key constraint may not exist
                     cursor.execute(stmt, (constraint['refpk'], ))
-                    foreign_table = cursor.fetchone()[0]
-                    constraint['reftable'] = foreign_table
-                    stmt = """SELECT column_name
-                              FROM   user_cons_columns
-                              WHERE  table_name = :table_name
-                              AND    constraint_name = :constraint_name
-                              ORDER BY position""" 
-                    cursor.execute(stmt, (constraint['reftable'], constraint['refpk']))
-                    columns = []
-                    for column in cursor.fetchall():
-                        columns.append(column[0])
-                    constraint['refcolumns'] = columns
+                    try:
+                        foreign_table = cursor.fetchone()[0]
+                    except TypeError:
+                        LOG.info('Cannot resolve foreign key constraint %s' %
+                                constraint['refpk'])
+                        foreign_table = None
+                    if foreign_table:
+                        constraint['reftable'] = foreign_table
+                        stmt = """SELECT column_name
+                                  FROM   all_cons_columns
+                                  WHERE  table_name = :table_name
+                                  AND    constraint_name = :constraint_name
+                                  ORDER BY position""" 
+                        cursor.execute(stmt, (constraint['reftable'], constraint['refpk']))
+                        columns = []
+                        for column in cursor.fetchall():
+                            columns.append(column[0])
+                        constraint['refcolumns'] = columns
             elif constraint['type'] == 'Check':
                 if not constraint['name'].startswith("SYS_C"):
                     constraint['condition'] = key[3]
             self.constraints[constraint['name']] = constraint
         # Indexes
         stmt = """SELECT index_name, index_type, uniqueness
-                  FROM   user_indexes
-                  WHERE  table_name = :table_name""" 
-        cursor.execute(stmt, (uc_table_name, ))
+                  FROM   all_indexes
+                  WHERE  table_name = :table_name
+                  AND    owner = :schema_name""" 
+        cursor.execute(stmt, (uc_table_name, self.schema))
         LOG.debug('Getting indexes for table %s' % uc_table_name)
         for index in cursor.fetchall():
             index_name = index[0]
@@ -316,20 +417,22 @@ class Table(schema.Table, OracleCalcPrecisionMixin):
             else:
                 index_dict['unique'] = False
             stmt = """SELECT column_name
-                      FROM   user_ind_columns
+                      FROM   all_ind_columns
                       WHERE  table_name = :table_name
+                      AND    table_owner = :schema_name
                       AND    index_name = :index_name
                       ORDER BY column_position""" 
-            cursor.execute(stmt, (uc_table_name, index_name))
+            cursor.execute(stmt, (uc_table_name, self.schema, index_name))
             index_dict['columns'] = []
             for column in cursor.fetchall():
                 index_dict['columns'].append(column[0])
             self.indexes[index_name] = index_dict
         # Triggers
         stmt = """SELECT trigger_name
-                  FROM   user_triggers
-                  WHERE  table_name = :table_name""" 
-        cursor.execute(stmt, (uc_table_name, ))
+                  FROM   all_triggers
+                  WHERE  table_name = :table_name
+                  AND    owner = :schema_name""" 
+        cursor.execute(stmt, (uc_table_name, self.schema))
         LOG.debug('Getting triggers for table %s' % uc_table_name)
         for trigger in cursor.fetchall():
             # Trigger name is the first element 
@@ -505,7 +608,10 @@ class View(schema.View, OracleCalcPrecisionMixin):
     """
 
     view_sql = {
-      'text': "SELECT text FROM user_views WHERE view_name=:name",
+      'text': """SELECT text 
+                 FROM all_views 
+                 WHERE view_name=:name 
+                 AND owner=:schema_name""",
       'columns': """SELECT column_id, column_name, data_type, data_length,
                            data_precision, data_scale, nullable
                     FROM   user_tab_columns
@@ -523,13 +629,25 @@ class View(schema.View, OracleCalcPrecisionMixin):
           cursor.
         @type cursor: Database cursor object
         """
-        cursor.execute(self.view_sql['text'], (self.name, ))
+        cursor.execute(self.view_sql['text'], (self.name, self.schema))
         LOG.debug('Getting text for index %s' % self.name)
         self.sql = cursor.fetchone()[0]
         # Columns, bizarrely these are stored in user_tab_columns
         cursor.execute(self.view_sql['columns'], (self.name, ))
         LOG.debug('Getting columns for index %s' % self.name)
-        self.columns = cursor.fetchall()
+        for col_row in cursor.fetchall():
+            col_name = col_row[1]
+            column = { 'sequence': col_row[0], 'name': col_name }
+            column['type'] = col_row[2]
+            column['length'] = col_row[3]
+            column['precision'] = col_row[4]
+            column['scale'] = col_row[5]
+            if col_row[6] == 'Y':
+                column['nullable'] = True
+            else:
+                column['nullable'] = False
+            self.columns[col_name] = column
+
         # Triggers
         cursor.execute(self.view_sql['triggers'], (self.name, ))
         LOG.debug('Getting triggers for index %s' % self.name)
@@ -546,19 +664,19 @@ class View(schema.View, OracleCalcPrecisionMixin):
         """
         if not hasattr(self, 'name') or self.name == None:
             raise AttributeError, "Can't generate DDL for a view without a name"
-        ddl_strings = ["CREATE VIEW "+self.name]
+        ddl_strings = ["CREATE VIEW %s" % self.name]
         in_columns = False
         # The columns may not necessarily be in the correct order. So we make a 
         # copy and sort by the first element of the tuple - the column number
-        sorted_columns = self.columns[:]
-        sorted_columns.sort()
-        for column in sorted_columns:
+        cols = self.columns.values()
+        cols.sort()
+        for column in cols:
             if in_columns:
                 ddl_strings.append("\n ,")
             else:
                 ddl_strings.append(" ( ")
                 in_columns = True
-            ddl_strings.append(column[1]+" ")
+            ddl_strings.append(column['name']+" ")
         # Only close the columns clause if we've actually got any columns
         if in_columns:
             ddl_strings.append(") AS\n  ")
@@ -588,7 +706,7 @@ class Trigger(schema.Trigger):
         """
         stmt = """SELECT trigger_type, triggering_event, 
                          trigger_body, table_name
-                  FROM   user_triggers
+                  FROM   all_triggers
                   WHERE  trigger_name = :name""" 
         LOG.debug('Getting details for trigger %s' % self.name)
         cursor.execute(stmt, (self.name, ))
@@ -636,20 +754,24 @@ class Sequence(schema.Sequence):
     """
     def _get_sequence(self, cursor):
         """
-        Query the data dictionary for this sequence
+        Query the data dictionary for this sequence. 
+
+        Until Oracle provides a way to determine the current value for a
+        sequence without altering it curr_value will always be None.
 
         @param cursor: All of the select statements will be executed using this
           cursor.
         @type cursor: Database cursor object
         """
-        stmt = """SELECT min_value, max_value, increment_by, %s.nextval
-                  FROM   user_sequences
-                  WHERE  sequence_name=:seqname""" % self.name
+        stmt = """SELECT min_value, max_value, increment_by
+                  FROM   all_sequences
+                  WHERE  sequence_name=:seqname
+                  AND    sequence_owner=:schema"""
         LOG.debug('Getting details for sequence %s' % self.name)
         # Ensure sequence name is upper case
-        cursor.execute(stmt, seqname=self.name.upper()) 
+        cursor.execute(stmt, (self.name.upper(), self.schema))
         results = cursor.fetchone()
-        self.min_value, self.max_value, self.increment_by, self.curr_value = results
+        self.min_value, self.max_value, self.increment_by = results
 
     def get_ddl(self):
         """Produce the DDL necessary to create this sequence
@@ -694,11 +816,12 @@ class CodeObject(schema.CodeObject):
         @type cursor: Database cursor object
         """
         stmt = """SELECT line, text 
-                  FROM   user_source 
+                  FROM   all_source 
                   WHERE  name=:object_name
                   AND    type=:object_type
+                  AND    owner=:schema
                   ORDER BY line"""
-        cursor.execute(stmt, (self.name, self.type))
+        cursor.execute(stmt, (self.name, self.type, self.schema))
         LOG.debug('Getting source for code objects %s' % self.name)
         self.source = []
         for code_line in cursor.fetchall():
